;;; internal.el --- The heart of internal version of Aldor Scanner

;; Author: Jeremy (Ya-Yu) Hsieh <hsiehy@unbc.ca>
;; Maintainer: None

;; This file contains the code for internal version Scanner
;; for Aldor


;;; Commentary:

;; This file contains the code and implementation for the internal
;; version of Aldor Scanner.  The main challenge while I am working
;; on this project is the rules.  Aldor has a very uniqe keyword
;; definition.  For example, 0 and 1 can be treated as identifier
;; integer, float, boolean, etc..  The float point type number
;; is very confusing too.  In this approach, I had two versions
;; 1) write to intermediate buffer and output; 2) change directly,
;; changed the buffer as it read alone.  I took the second approach.
;; since it is faster and simpler.

;; For more detail of the logic and information, please refer
;; to my documentation/report/thesis



;; Some basic regular expressions 
(setq esc "_"
      alpha '(any "a-zA-Z")
      one-digit '(any "2-9")

; Already built-in       (digit (rx (any "0-9")))

      esc-digit '(and (zero-or-one "_") digit)
      long-digit '(any "0-9A-Za-z")   ; CAN be replace by "letter"....
      esc-long-digit `(and (zero-or-one "_") ,long-digit)
      radix `(and digit (0+ ,esc-digit) "r")
      expon `(and (any "eE")
		   (zero-or-one (any "+-"))
		   (1+ ,esc-digit)))

;; NON-ALDOR WORDS - SPACE, TAB, NEWLINE, etc.
(setq scanner-internal-other (rx (or
		 blank
		 space
		 "\n"
		 "\015")))
      
;; RESERVED WORDS
(setq aldor-reserved-word (rx (and 
	      word-start
	      (or
	      "add"        "always"    "and"         "assert"    "break"     
	      "but"        "catch"     "default"     "define"    "delay"     
	      "do"         "else"      "except"      "export"    "extend"  
	      "fix"        "fluid"     "for"         "free"      "from"      
	      "generate"   "goto"      "has"         "if"        "import" 
	      "in"         "inline"    "is"          "isnt"      "iterate"  
	      "let"        "local"     "macro"       "never"     "not"
	      "of"         "or"        "pretend"     "ref"       "repeat"  
	      "return"     "rule"      "select"      "then"      "throw"   
	      "to"         "try"       "where"       "while"     "with"  
	      "yield")
	      word-end)))


;; DEFINABLE KEYWORD
(setq aldor-definable-op-word (rx (and 
	      word-start
	      (or   "by"      "case"    "mod"    "quo"    "rem") 
	      word-end)))


;; CLASS
(setq aldor-class (rx (and 
	      word-start
	      (or
	      "boolean"    "DoubleFloat"    "false"      "Integer"
	      "List"       "TextWriter"     "String"     "true"
	      "Character"  "MachineInteger")
	      word-end)))


;; IMPORT FUNCTION
(setq aldor-import-function (rx (and 
	      word-start
	      (or
	      "stdout"     "newline"    "space"      "include"    "rep"
	      "per"        "Per")
	      word-end)))


;; PRE-DOCUMENT
(setq aldor-pre-document (rx (and (group (repeat 3 "+"))
			    (1+ not-newline))))


;; POST-DOCUMENT
(setq aldor-post-document (rx (and (group (repeat 2 "+"))
			     (1+ not-newline))))


;; COMMENT
(setq aldor-comment (rx (and (group (repeat 2 "-"))
		       (1+ not-newline))))


;; STRING 
;string  \"[^\"]*\"
(setq aldor-string (rx (and "\"" 
		      (0+ (or (and "_" (1+ space))
			      (and "_" (not space))
			      (not (any "_\n\""))))
     		      "\"" )))

;; IDENTIFIER - NOT RESERVED WORDS
;id  "0"|"1"|"%"|{alpha}({alpha}|{digit}|[!?|])*
(setq aldor-id (rx (and 
	      word-start
	      (or (and "0" word-end)
		  (and "1" word-end)
		  (and (any "a-zA-Z")
		       (* (or (any "a-zA-Z0-9!?|")))))
	      )))

;; FLOAT
; 
; 
;float    {digit}*"."{esc-digit}+{expon}?|
;         {digit}+"."{esc-digit}*{expon}?|
;         {digit}+{expon}|
;         {radix}{esc-long-digit}*"."{esc-long-digit}+{expon}?|
;         {radix}{esc-long-digit}+"."{esc-long-digit}*{expon}?|
;         {radix}{esc-long-digit}+{expon}

(setq aldor-float (rx-to-string `(or
			    (and (* digit) "." 
				 (+ ,esc-digit) (zero-or-one ,expon))
			    (and digit (* ,esc-digit) "." (* ,esc-digit) 
				 (zero-or-one ,expon))
			    (and digit (* ,esc-digit) ,expon)
			    (and ,radix
				 (or 
				  (and (* ,esc-long-digit) 
				       "."
				       (+ ,esc-long-digit)
				       (zero-or-more ,expon))
				  (and (+ ,esc-long-digit)
				       "."
				       (* ,esc-long-digit)
				       (zero-or-more ,expon))
				  (and (+ ,esc-long-digit) ,expon))))))
			    


;; INTEGER
; 
; 
;int  {one-digit}|{digit}{esc-digit}+|{radix}{esc-long-digit}+
(setq aldor-integer (rx-to-string `(or
			      ,one-digit
			      (and digit (1+ ,esc-digit))
			      (and ,radix (1+ ,esc-long-digit)))))


;; DEFINABLE OPERATOR
(setq aldor-definable-op (rx (or
			"\\"  "#"   "*"   "+"    "-"
			"/"   "<"   "="   ">"    "@"
			"~"   "^"   "<<"  ">>"   "<="
			">="  "<-"  "->"  "\\/"  "/\\"
			"**"  ".."  "~="  "^=")))
			 

;; RESERVED OPERATOR
(setq aldor-reserved-op (rx (or
		       "$"    "`"   "("    ")"   "+->"
		       ","    ":"   ":="   ";"   "==>"
		       "=>"   "["   "]"    "{"   "}"
		       "=="   "|")))

;; FUTURE OPERATOR
(setq aldor-future-op (rx (or
		     "[|"   "{|"    "(|"   "|]"
		     "|}"   "|)"    "`"    "&"
		     "||")))





;; The main program to call, which should do ALL the works
;;   - add the path
;;   - set all the variables (source buffer, result buffer, and process)
;;   - called "find-token" to read from the the source file and write
;;     into the buffer.

(defun aldor-scan (str)
  (interactive "fFile to scan? ")
  (let
      ((exec-path (cons default-directory exec-path))
       (result (find-file-noselect str)))
    (aldor-find-token result)))



(defun aldor-scan-buffer ()
  (interactive)
  (let
      ((exec-path (cons default-directory exec-path))
       (result (find-file-noselect buffer-file-name)))
    (aldor-find-token result)))


;; "find-token" take the source buffer (result) and process it with the
;; internal scanner, then write the result back into the source buffer.

(defun aldor-find-token (result)
  (save-excursion
    (let ((start 1))
      (set-buffer result)
      (beginning-of-buffer)
      (while (not (eobp))
	(cond
	 ((looking-at aldor-pre-document)
	  (aldor-colour-syntax start (re-search-forward aldor-pre-document) 'aldor-pre-doc result))
	 ((looking-at aldor-post-document)
	  (aldor-colour-syntax start (re-search-forward aldor-post-document) 'aldor-post-doc result))
	 ((looking-at aldor-comment) 
	  (aldor-colour-syntax start (re-search-forward aldor-comment) 'aldor-comment result))
	 ((looking-at aldor-string) 
	  (aldor-colour-syntax start (re-search-forward aldor-string) 'aldor-string result))
	 ((looking-at aldor-reserved-word)
	  (aldor-colour-syntax start (re-search-forward aldor-reserved-word) 'aldor-reserved result))
	 ((looking-at aldor-definable-op-word) 
	  (aldor-colour-syntax start (re-search-forward aldor-definable-op-word) 'aldor-def-op-word result))
	 ((looking-at aldor-class) 
	  (aldor-colour-syntax start (re-search-forward aldor-class) 'aldor-class result))
	 ((looking-at aldor-import-function)
	  (aldor-colour-syntax start (re-search-forward aldor-import-function) 'aldor-import-function result))
	 ((looking-at aldor-definable-op) 
	  (aldor-colour-syntax start (re-search-forward aldor-definable-op) 'aldor-def-op result))
	 ((looking-at aldor-reserved-op) 
	  (aldor-colour-syntax start (re-search-forward aldor-reserved-op) 'aldor-res-op result))
	 ((looking-at aldor-future-op)
	  (aldor-colour-syntax start (re-search-forward aldor-future-op) 'aldor-future-op result))
	 ((looking-at aldor-float) 
	  (aldor-colour-syntax start (re-search-forward aldor-float) 'aldor-float result))
	 ((looking-at aldor-integer)
	  (aldor-colour-syntax start (re-search-forward aldor-integer) 'aldor-int result))
	 ((looking-at aldor-id) 
	  (aldor-colour-syntax start (re-search-forward aldor-id) 'aldor-id result))
	 ((looking-at scanner-internal-other)
	  (re-search-forward scanner-internal-other)))
	(setq start (point)))))
)


;; "aldor-colour-syntax" function does the face colour change for the result
;; buffer.  This function takes four arguments -
;;    - start: The start point of the token
;;    - end  : The end point of the token
;;    - type : Type of the token, so it can do the different colour change
;;    - result: The buffer where the colour should be apply to
;;
;; Note:
;;    - This function change the buffer colour ONLY.  Therefore, this 
;;      function does NOT save the file.  Therefore, if the user
;;      does not save the buffer, then the change will be erase.
;;      (Primary goal is to avoid change the user's file)
;;    - Set aldor-def-op-word to the same face - keyword face
;;    - Set aldor-pre-doc and aldor-post-doc has same color - doc face
;;    - Set float and int has same color - constant face
;;    - Set operators (def, res, & future)to  same color - buildin face

(defun aldor-colour-syntax( start end type result)
      (cond ((equal type 'aldor-reserved)
	     (put-text-property start end 'face 'font-lock-keyword-face result)
	     (put-text-property start end 'help-echo "RESERVED" result))
	    ((equal type 'aldor-def-op-word)
	     (put-text-property start end 'face 'font-lock-keyword-face result)
	     (put-text-property start end 'help-echo "DEFINABLE KEYWORD" result))
	    ((equal type 'aldor-class)
	     (put-text-property start end 'face 'font-lock-type-face result)
	     (put-text-property start end 'help-echo "CLASS" result))
	    ((equal type 'aldor-import-function)
	     (put-text-property start end 'face 'font-lock-function-name-face result)
	     (put-text-property start end 'help-echo "IMPORT FUNCTION" result))
	    ((equal type 'aldor-pre-doc)
	     (put-text-property start end 'face 'font-lock-doc-face result)
	     (put-text-property start end 'help-echo "PRE-DOCUMENT" result))
	    ((equal type 'aldor-post-doc)
	     (put-text-property start end 'face 'font-lock-doc-face result)
	     (put-text-property start end 'help-echo "POST-DOCUMENT" result))
	    ((equal type 'aldor-comment)
	     (put-text-property start end 'face 'font-lock-comment-face result)
	     (put-text-property start end 'help-echo "COMMENT" result))
	    ((equal type 'aldor-string)
	     (put-text-property start end 'face 'font-lock-string-face result)
	     (put-text-property start end 'help-echo "STRING" result))
	    ((equal type 'aldor-id)
	     (put-text-property start end 'face 'font-lock-variable-name-face result)
	     (put-text-property start end 'help-echo "IDENTIFIER" result))
	    ((equal type 'aldor-float)
	     (put-text-property start end 'face 'font-lock-constant-face result)
	     (put-text-property start end 'help-echo "FLOAT" result))
	    ((equal type 'aldor-int)
	     (put-text-property start end 'face 'font-lock-constant-face result)
	     (put-text-property start end 'help-echo "INTEGER" result))
	    ((equal type 'aldor-def-op)
	     (put-text-property start end 'face 'font-lock-builtin-face result)
	     (put-text-property start end 'help-echo "DEFINABLE OPERATOR" result))
	    ((equal type 'aldor-res-op)
	     (put-text-property start end 'face 'font-lock-builtin-face result)
	     (put-text-property start end 'help-echo "RESERVED OPERATOR" result))
	    ((equal type 'aldor-future-op)
	     (put-text-property start end 'face 'font-lock-buildin-face result)
	     (put-text-property start end 'help-echo "FUTURE OPERATOR" result))))

(provide 'aldor-internal)