;;; internal.el --- The heart of internal version of Aldor Scanner

;; Author: Jeremy (Ya-Yu) Hsieh <hsiehy@unbc.ca>
;; Maintainer: None

;; This file contains the code for internal version Scanner
;; for Aldor


;;; Commentary:

;; This file contains the code and implementation for the internal
;; version of Aldor Scanner.  The main challenge while I am working
;; on this project is the rules.  Aldor has a very uniqe keyword
;; definition.  For example, 0 and 1 can be treated as identifier
;; integer, float, boolean, etc..  The float point type number
;; is very confusing too.  In this approach, I had two versions
;; 1) write to intermediate buffer and output; 2) change directly,
;; changed the buffer as it read alone.  I took the second approach.
;; since it is faster and simpler.

;; For more detail of the logic and information, please refer
;; to my documentation/report/thesis



;; Some basic regular expressions
(setq aldor-esc "_"
      aldor-alpha '(any "a-zA-Z")
      aldor-one-digit '(any "2-9")

; Already built-in       (digit (rx (any "0-9")))

      aldor-esc-digit '(and (zero-or-one "_") digit)
      aldor-long-digit '(any "0-9A-Za-z")   ; CAN be replace by "letter"....
      aldor-esc-long-digit `(and (zero-or-one "_") ,aldor-long-digit)
      aldor-radix `(and digit (0+ ,aldor-esc-digit) "r")
      aldor-expon `(and (any "eE")
		   (zero-or-one (any "+-"))
		   (1+ ,aldor-esc-digit)))

;; NON-ALDOR WORDS - SPACE, TAB, NEWLINE, etc.
(setq aldor-other (rx (or
		 blank
		 space
		 "\n"
		 "\015")))
      
;; RESERVED WORDS
(setq aldor-reserved-word (rx (and 
	      word-start
	      (or
	      "add"        "always"    "and"         "assert"    "break"     
	      "but"        "catch"     "default"     "define"    "delay"     
	      "do"         "else"      "except"      "export"    "extend"  
	      "fix"        "fluid"     "for"         "free"      "from"      
	      "generate"   "goto"      "has"         "if"        "import" 
	      "in"         "inline"    "is"          "isnt"      "iterate"  
	      "let"        "local"     "macro"       "never"     "not"
	      "of"         "or"        "pretend"     "ref"       "repeat"  
	      "return"     "rule"      "select"      "then"      "throw"   
	      "to"         "try"       "where"       "while"     "with"  
	      "yield")
	      word-end)))


;; DEFINABLE KEYWORD
(setq aldor-definable-op-word (rx (and 
	      word-start
	      (or   "by"      "case"    "mod"    "quo"    "rem") 
	      word-end)))


;; CLASS
(setq aldor-class (rx (and 
	      word-start
	      (or
	      "boolean"    "DoubleFloat"    "false"      "Integer"
	      "List"       "TextWriter"     "String"     "true"
	      "Character"  "MachineInteger")
	      word-end)))


;; IMPORT FUNCTION
(setq aldor-import-function (rx (and 
	      word-start
	      (or
	      "stdout"     "newline"    "space"      "include"    "rep"
	      "per"        "Per")
	      word-end)))


;; PRE-DOCUMENT
(setq aldor-pre-document (rx (and (group (repeat 3 "+"))
			    (1+ not-newline))))


;; POST-DOCUMENT
(setq aldor-post-document (rx (and (group (repeat 2 "+"))
			     (1+ not-newline))))


;; COMMENT
(setq aldor-comment (rx (and (group (repeat 2 "-"))
		       (1+ not-newline))))


;; STRING
;string  \"[^\"]*\"
(setq aldor-string (rx (and "\"" 
		      (0+ (or (and "_" (1+ space))
			      (and "_" (not space))
			      (not (any "_\n\""))))
     		      "\"" )))

;; IDENTIFIER - NOT RESERVED WORDS
;id  "0"|"1"|"%"|{aldor-alpha}({aldor-alpha}|{digit}|[!?|])*
(setq aldor-id (rx (and 
	      word-start
	      (or (and "0" word-end)
		  (and "1" word-end)
		  (and (any "a-zA-Z")
		       (* (or (any "a-zA-Z0-9!?|")))))
	      )))

;; FLOAT
;
;
;float    {digit}*"."{aldor-esc-digit}+{expon}?|
;         {digit}+"."{aldor-esc-digit}*{expon}?|
;         {digit}+{expon}|
;         {radix}{aldor-esc-long-digit}*"."{aldor-esc-long-digit}+{expon}?|
;         {radix}{aldor-esc-long-digit}+"."{aldor-esc-long-digit}*{expon}?|
;         {radix}{aldor-esc-long-digit}+{expon}

(setq aldor-float (rx-to-string `(or
			    (and (* digit) "." 
				 (+ ,aldor-esc-digit) (zero-or-one ,aldor-expon))
			    (and digit (* ,aldor-esc-digit) "." (* ,aldor-esc-digit) 
				 (zero-or-one ,aldor-expon))
			    (and digit (* ,aldor-esc-digit) ,aldor-expon)
			    (and ,aldor-radix
				 (or 
				  (and (* ,aldor-esc-long-digit) 
				       "."
				       (+ ,aldor-esc-long-digit)
				       (zero-or-more ,aldor-expon))
				  (and (+ ,aldor-esc-long-digit)
				       "."
				       (* ,aldor-esc-long-digit)
				       (zero-or-more ,aldor-expon))
				  (and (+ ,aldor-esc-long-digit) ,aldor-expon))))))
			    


;; INTEGER
;
;
;int  {one-digit}|{digit}{aldor-esc-digit}+|{radix}{aldor-esc-long-digit}+
(setq aldor-integer (rx-to-string `(or
			      ,aldor-one-digit
			      (and digit (1+ ,aldor-esc-digit))
			      (and ,aldor-radix (1+ ,aldor-esc-long-digit)))))


;; DEFINABLE OPERATOR
(setq aldor-definable-op (rx (or
			"\\"  "#"   "*"   "+"    "-"
			"/"   "<"   "="   ">"    "@"
			"~"   "^"   "<<"  ">>"   "<="
			">="  "<-"  "->"  "\\/"  "/\\"
			"**"  ".."  "~="  "^=")))
			 

;; RESERVED OPERATOR
(setq aldor-reserved-op (rx (or
		       "$"    "`"   "("    ")"   "+->"
		       ","    ":"   ":="   ";"   "==>"
		       "=>"   "["   "]"    "{"   "}"
		       "=="   "|")))

;; FUTURE OPERATOR
(setq aldor-future-op (rx (or
		     "[|"   "{|"    "(|"   "|]"
		     "|}"   "|)"    "`"    "&"
		     "||")))





;; The main program to call, which should do ALL the works
;;   - add the path
;;   - set all the variables (source buffer, result buffer, and process)
;;   - called "aldor-region-find-token" to read from the the source file and write
;;     into the buffer.

(defun aldor-scan-this-buffer ()
  (interactive)
  (aldor-region-find-token (current-buffer) (region-beginning) (region-end)))

(defvar my-beg nil)
(defvar my-end nil)

;; NOTE: "\t" somehow doesn't work;  ";" will NOT be trigger, since
;;       it always become part of the others.
;;       Maybe " " does not require as well.  Since eventually the
;;       programmer will hit the newline ("\n") character anway.  
(defvar triggers '("\n", " "))

(defun aldor-scan-region( beg end )
  (interactive)
  (setq my-beg (if my-beg (min my-beg beg)
		 beg)
	my-end (if my-end (max my-end end) end))
  (if ;(string= (buffer-substring-no-properties beg end) "\n")
      (member (buffer-substring-no-properties beg end) triggers)
      (progn
	(let ((inhibit-modification-hooks t)
	      (begin (save-excursion
		       (goto-char my-beg)
		       (line-beginning-position)))
	      (finish (save-excursion
			(goto-char my-end)
			(line-end-position 2))))
	  (aldor-region-find-token (current-buffer) begin finish)
	  (setq my-end nil my-beg nil)))))


;; "aldor-region-find-token" take the source buffer (result) and process it with the
;; internal scanner, then write the result back into the source buffer.
;; use "narrow-to-region" to narrow down the buffer, limit the access
;; level for the scanner.  At the end, use "widen" to bring everything
;; back.

;; NOTE: ATTENTION!! CAUTION!!
;;         EVEN WITH THE NARROW, THE POINT STILL HAS THE SAME VALUE,
;;         IT WILL NOT START AT 1.

(defvar aldor-face-regexps
  '(( aldor-pre-document . aldor-pre-doc)
    ( aldor-post-document . aldor-post-doc)
    ( aldor-comment . aldor-comment)
    ( aldor-string . aldor-string)
    ( aldor-reserved-word . aldor-reserved)
    ( aldor-definable-op-word . aldor-def-op-word)
    ( aldor-class . aldor-class)
    ( aldor-import-function . aldor-import-function)
    ( aldor-definable-op . aldor-def-op)
    ( aldor-reserved-op . aldor-res-op)
    ( aldor-future-op . aldor-future-op)
    ( aldor-float . aldor-float)
    ( aldor-integer . aldor-int)
    ( aldor-id . aldor-id)))


(defun aldor-region-find-token (result first last)
  (save-excursion
    (let ((start 0) regexp face regexp-list colored)
      (set-buffer result)
      (goto-char first)
      (setq start first)
      (while (< (point) last)
	(setq regexp-list aldor-face-regexps
	      colored nil)
	(while (not (null regexp-list))
	  (setq face (cdr (car regexp-list))
		regexp (car (car regexp-list))
		regexp-list (cdr regexp-list))
	  (if (looking-at (eval regexp))
	      (progn
		(setq colored t)
		(aldor-colour-syntax start (match-end 0) face result)
		(setq regexp-list nil))))
	(and (not colored) (looking-at aldor-other))
	(goto-char (match-end 0))
	(setq start (point))))))

;; eval: to evaluate....... to a string




;; "aldor-colour-syntax" function does the face colour change for the result
;; buffer.  This function takes four arguments -
;;    - start: The start point of the token
;;    - end  : The end point of the token
;;    - type : Type of the token, so it can do the different colour change
;;    - result: The buffer where the colour should be apply to
;;
;; Note:
;;    - This function change the buffer colour ONLY.  Therefore, this 
;;      function does NOT save the file.  Therefore, if the user
;;      does not save the buffer, then the change will be erase.
;;      (Primary goal is to avoid change the user's file)
;;    - Set aldor-def-op-word to the same face - keyword face
;;    - Set aldor-pre-doc and aldor-post-doc has same color - doc face
;;    - Set float and int has same color - constant face
;;    - Set operators (def, res, & future)to  same color - buildin face

(defun aldor-colour-syntax( start end type result)
  (cond ((equal type 'aldor-reserved)
	 (put-text-property start end 'face 'font-lock-keyword-face result)
	 (put-text-property start end 'help-echo "RESERVED" result))
	((equal type 'aldor-def-op-word)
	 (put-text-property start end 'face 'font-lock-keyword-face result)
	 (put-text-property start end 'help-echo "DEFINABLE KEYWORD" result))
	((equal type 'aldor-class)
	 (put-text-property start end 'face 'font-lock-type-face result)
	 (put-text-property start end 'help-echo "CLASS" result))
	((equal type 'aldor-import-function)
	 (put-text-property start end 'face 'font-lock-function-name-face result)
	 (put-text-property start end 'help-echo "IMPORT FUNCTION" result))
	((equal type 'aldor-pre-doc)
	 (put-text-property start end 'face 'font-lock-doc-face result)
	 (put-text-property start end 'help-echo "PRE-DOCUMENT" result))
	((equal type 'aldor-post-doc)
	 (put-text-property start end 'face 'font-lock-doc-face result)
	 (put-text-property start end 'help-echo "POST-DOCUMENT" result))
	((equal type 'aldor-comment)
	 (put-text-property start end 'face 'font-lock-comment-face result)
	 (put-text-property start end 'help-echo "COMMENT" result))
	((equal type 'aldor-string)
	 (put-text-property start end 'face 'font-lock-string-face result)
	 (put-text-property start end 'help-echo "STRING" result))
	((equal type 'aldor-id)
	 (put-text-property start end 'face 'font-lock-variable-name-face result)
	 (put-text-property start end 'help-echo "IDENTIFIER" result))
	((equal type 'aldor-float)
	 (put-text-property start end 'face 'font-lock-constant-face result)
	 (put-text-property start end 'help-echo "FLOAT" result))
	((equal type 'aldor-int)
	 (put-text-property start end 'face 'font-lock-constant-face result)
	 (put-text-property start end 'help-echo "INTEGER" result))
	((equal type 'aldor-def-op)
	 (put-text-property start end 'face 'font-lock-builtin-face result)
	 (put-text-property start end 'help-echo "DEFINABLE OPERATOR" result))
	((equal type 'aldor-res-op)
	 (put-text-property start end 'face 'font-lock-builtin-face result)
	 (put-text-property start end 'help-echo "RESERVED OPERATOR" result))
	((equal type 'aldor-future-op)
	 (put-text-property start end 'face 'font-lock-buildin-face result)
	 (put-text-property start end 'help-echo "FUTURE OPERATOR" result))))

(provide 'aldor-region-in)